class NoCasesError(Exception):
    """Raised when no cases are found in the database."""
    pass


class NoValidAnswers(Exception):
    pass


class NoSchoolIdsError(Exception):
    pass


class NoAnswersError(Exception):
    pass
